class DynamicBatchIndexError(IndexError):
    """
    Error raised when the number of outputs returned by the infer function of a InferenceModel
    differs from the number of inputs.
    """

    def __init__(self, message):
        """
        The DynamicBatchIndexError is used to propagate the cause of the Errors to users of the DynBatcher.

        :param message: Message to be propagated back
        :type message: str, optional
        """
        super().__init__(message)
        self.message = message

    def __str__(self):
        return f"DynamicBatchIndexError: {self.message}"



import typing as t

from pydantic import BaseModel


class Task(BaseModel):
    """A model of a task used by the DynBatcher. It is used to return the batched inputs back to the right calling thread."""

    id: int
    content: t.Any



import asyncio

import pytest

from fast_dynamic_batcher.errors import DynamicBatchIndexError


@pytest.mark.asyncio
async def test_raised_errors_not_returned(dyn_batcher):
    await_list = []
    for i in range(5):
        await_list.append(dyn_batcher.process_batched(i))
    with pytest.raises(DynamicBatchIndexError) as exc_info:
        await asyncio.gather(*await_list, return_exceptions=False)
    assert exc_info.type is DynamicBatchIndexError
    assert exc_info.value.message == "Dynamic batch of input size 5 produced 4."


@pytest.mark.asyncio
async def test_raised_errors_returned(dyn_batcher):
    await_list = []
    for i in range(5):
        await_list.append(dyn_batcher.process_batched(i))
    results = await asyncio.gather(*await_list, return_exceptions=True)

    assert all([isinstance(res, DynamicBatchIndexError) for res in results])
    assert all([res.message == "Dynamic batch of input size 5 produced 4." for res in results])



from abc import ABC, abstractmethod
from typing import Any


class InferenceModel(ABC):
    """
    The InferenceModel abstract class is used as a template in the DynBatcher. An implementation of it must be passed to it.
    The __main__ method of an implementation is used to load the machine learning model in the worker process created by the DynBatcher.
    The infer method is used to process the batched inputs.

    :param ABC: Dynamic base class
    """

    @abstractmethod
    def infer(self, inputs: list[Any]) -> list[Any]:
        """
        Abstract method that is used by the DynBatcher to process a batch of inputs.


        :param inputs: A list of inputs your model can run as a single batch.
        :type inputs: list[Any]
        :return: The outputs of the machine learning model. The position of an output should correspond to the position of the input used to produce it.
        :rtype: list[Any]
        """
        ...



import asyncio
import logging
import multiprocessing as mp
import typing as t

from concurrent.futures import Future
from ctypes import c_bool
from threading import Event, Lock, Thread

from fast_dynamic_batcher.dyn_prog_loop import _main_loop, _return_loop
from fast_dynamic_batcher.errors import DynamicBatchIndexError
from fast_dynamic_batcher.inference_template import InferenceModel
from fast_dynamic_batcher.models import Task


try:
    mp.set_start_method("spawn", force=True)
except RuntimeError:
    pass


class DynBatcher:
    """The DynBatcher class is used to batch inputs across many requests into a single batch to accelerate ML workloads."""

    def __init__(
        self,
        inference_model: InferenceModel,
        max_batch_size: int = 8,
        max_delay: float = 0.1,
        queue_size: int = 10000,
    ):
        """
        Initializes the Dynamic Batcher.

        :param inference_model: Implementation of the abstract InferenceModel class
        :type inference_model: InferenceModel
        :param max_batch_size: The maximal number of inputs that are processed together, defaults to 8
        :type max_batch_size: int, optional
        :param max_delay: The maximal delay between receiving the first input and the start of the batch processing, defaults to 0.1
        :type max_delay: float, optional
        :param queue_size: The maximal size of the queue holding batched inputs, defaults to 10000
        :type queue_size: int, optional
        """
        # mp.set_start_method("spawn", force=True)  # otherwise torch does not work.
        self.logger = mp.get_logger()  # mp.log_to_stderr()
        self.logger.setLevel(logging.INFO)
        self._input_queue = mp.Queue(queue_size)
        self._output_queue = mp.Queue(queue_size)
        self._unique_id_counter = 0
        self._state_map: dict[int, Future] = {}
        self._state_lock = Lock()
        self._lock = asyncio.Lock()
        self.shared_bool = mp.Value(c_bool, True)  # TODO: Check if RawValue is faster because of no lock
        self._batcher = mp.Process(
            target=_main_loop,
            args=(inference_model, self._input_queue, self._output_queue, self.shared_bool, max_batch_size, max_delay),
            daemon=True,
        )

        # TODO: enable creation as singleton
        self._batcher.start()
        self.logger.info("Batcher process started")

        # TODO: create a coroutine / Task that runs endlessly and processes the output queue
        # Maybe with asyncio.to_thread (just runs it in a different thread, still needs await)
        # loop.create_task()
        self.event = Event()
        self.thread = Thread(
            target=_return_loop, args=(self._state_map, self._state_lock, self._output_queue, self.event)
        )
        self.thread.start()

        self.logger.info("Output thread started")

    def stop(self):
        """
        Stops all processes and threads created by the DynBatcher.
        """
        self._input_queue.close()
        self._output_queue.close()
        self.event.set()  # joins self.thread
        self.shared_bool.value = False
        self._batcher.join()
        self.thread.join()
        # TODO: Terminate if join not working

    async def process_batched(self, input: t.Any) -> t.Any:
        """
        Process an input as a batch.

        :param input: The input which will be passed to the `infer` of the registered InferenceModel as the content of a Task.
        :type input: t.Any
        :raises DynamicBatchIndexError: Error raised when something went wrong during batch processing.
        :return: The output of the InferenceModel's `infer` method.
        :rtype: t.Any
        """
        result = await asyncio.to_thread(self._process_batched, input)
        if isinstance(result, DynamicBatchIndexError):
            raise result
        return result

    def _process_batched(self, input: t.Any) -> t.Any:
        future = Future()
        # async with self._lock:
        task_id = self._unique_id_counter
        # self._state_lock.acquire_lock()  # Todo: is the lock actually needed
        self._state_map[self._unique_id_counter] = future
        # self._state_lock.release_lock()
        self._unique_id_counter += 1

        task = Task(id=task_id, content=input)
        self._input_queue.put(task, False)  # raises queue.Full
        # TODO: error handling
        future_result = future.result()
        return future_result



import time

from asyncio import Future
from multiprocessing import Queue, Value
from threading import Event, Lock

from fast_dynamic_batcher.errors import DynamicBatchIndexError
from fast_dynamic_batcher.inference_template import InferenceModel
from fast_dynamic_batcher.models import Task


# TODO: Bound exp. backoff
RETURN_LOOP_SLEEP = 0.001
MAIN_LOOP_SLEEP = 0.001

TICKS_IN_NANO = 1000000000


def _main_loop(
    inference_model: InferenceModel,
    input_queue: Queue,
    output_queue: Queue,
    shared_bool: Value,
    max_batch_size: int = 8,
    max_delay: float = 0.1,
):
    inference_model = inference_model()
    start_time = time.monotonic_ns()
    max_delay_ns = int(TICKS_IN_NANO * max_delay)
    batch_list = []

    # TODO: add flag to stop the loop
    while shared_bool.value:
        if len(batch_list) == 0:
            start_time = time.monotonic_ns()
        no_op = True
        if not input_queue.empty():
            no_op = False
            # should not block
            input = input_queue.get()
            batch_list.append(input)
        if len(batch_list) == max_batch_size or start_time + max_delay_ns < time.monotonic_ns():
            no_op = False
            inputs = [t.content for t in batch_list]
            outputs = inference_model.infer(inputs)
            if len(outputs) != len(inputs):
                output_tasks = [
                    Task(
                        id=batch_list[i].id,
                        content=DynamicBatchIndexError(
                            f"Dynamic batch of input size {len(batch_list)} produced {len(outputs)}."
                        ),
                    )
                    for i in range(len(batch_list))
                ]
            else:
                output_tasks = [Task(id=batch_list[i].id, content=outputs[i]) for i in range(len(batch_list))]
            for output in output_tasks:
                output_queue.put(output)
            batch_list = []
        if no_op:
            time.sleep(MAIN_LOOP_SLEEP)


def _return_loop(state_map: dict[int, Future], state_lock: Lock, output_queue: Queue, event: Event):
    while not event.is_set():
        while not output_queue.empty():
            # shuold never block
            output = output_queue.get()
            # TODO: error handling
            state_lock.acquire()
            output_future = state_map[output.id]
            del state_map[output.id]
            state_lock.release()
            output_future.set_result(output.content)
        time.sleep(RETURN_LOOP_SLEEP)



import asyncio
import json
import time

import pytest

from asgi_lifespan import LifespanManager
from httpx import ASGITransport, AsyncClient


LOCAL = False


# TODO: asyncio_mode=auto
@pytest.mark.flaky(retries=3, delay=1)
@pytest.mark.asyncio
async def test_one(wrapped_app):
    async with LifespanManager(wrapped_app) as manager:
        start_s = time.monotonic()
        async with AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as client:
            response = await client.post("http://test/predict", json={"text": "hi"})
        took_s = time.monotonic() - start_s
        if LOCAL:
            assert (took_s > 0.05) and (took_s < 0.17)
        assert response.status_code == 200
        assert json.loads(response.json()) == {"text": "hi"}


@pytest.mark.flaky(retries=3, delay=1)
@pytest.mark.asyncio
async def test_eight(wrapped_app):
    async with LifespanManager(wrapped_app) as manager:
        start_s = time.monotonic()
        async with AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as client:
            requests = []
            for i in range(8):
                requests.append(client.post("http://test/predict", json={"text": f"{i}"}))
            results = await asyncio.gather(*requests)
        took_s = time.monotonic() - start_s
        if LOCAL:
            assert (took_s > 0.05) and (took_s < 0.2)
        assert all([response.status_code == 200 for response in results])


@pytest.mark.flaky(retries=3, delay=1)
@pytest.mark.asyncio
async def test_twelve(wrapped_app):
    async with LifespanManager(wrapped_app) as manager:
        start_s = time.monotonic()
        async with AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as client:
            requests = []
            for i in range(12):
                requests.append(client.post("http://test/predict", json={"text": f"{i}"}))
            results = await asyncio.gather(*requests)
        took_s = time.monotonic() - start_s
        if LOCAL:
            assert (took_s > 0.2) and (took_s < 0.3)
        assert all([response.status_code == 200 for response in results])


@pytest.mark.flaky(retries=3, delay=1)
@pytest.mark.asyncio
async def test_hundred(wrapped_app):
    async with LifespanManager(wrapped_app) as manager:
        start_s = time.monotonic()
        async with AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as client:
            requests = []
            for i in range(100):
                requests.append(client.post("http://test/predict", json={"text": f"{i}"}))
            results = await asyncio.gather(*requests)
        took_s = time.monotonic() - start_s
        print(took_s)
        if LOCAL:
            assert (took_s > 0.75) and (took_s < 0.9)
        assert all([response.status_code == 200 for response in results])
